@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005 Alex Murray <murray.alex@gmail.com>
+ * Copyright (C) 2005 Alex Murray <pragmatine@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -22,7 +22,7 @@
 #include "sensors-applet.h"
 
 void i2c_sys_sensors_interface_init(SensorsApplet *sensors_applet);
-gfloat i2c_sys_sensors_interface_get_sensor_value(const gchar *path, 
+gdouble i2c_sys_sensors_interface_get_sensor_value(const gchar *path, 
 						  const gchar *id, 
 						  SensorType type,
 						  GError **error);
