@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005 Alex Murray <murray.alex@gmail.com>
+ * Copyright (C) 2005 Alex Murray <pragmatine@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,6 +24,7 @@
 #include "config.h"
 #endif /* HAVE_CONFIG_H */
 
+#include <gnome.h>
 #include <glib/gprintf.h>
 #include "sensors-applet.h"
 #include "sensors-applet-gconf.h"
@@ -31,6 +32,7 @@
 #include "i2c-sys-sensors-interface.h"
 #include "i2c-proc-sensors-interface.h"
 #include "i8k-sensors-interface.h"
+#include "ibm-acpi-sensors-interface.h"
 #include "omnibook-sensors-interface.h"
 #include "pmu-sys-sensors-interface.h"
 #include "hddtemp-sensors-interface.h"
@@ -44,7 +46,11 @@
 static void prefs_cb(BonoboUIComponent *uic,
 		     SensorsApplet    *sensors_applet,
 		     const gchar       *verbname) {
-	prefs_dialog_show(sensors_applet);
+	if (sensors_applet->prefs_dialog) {
+		gtk_window_present(GTK_WINDOW(sensors_applet->prefs_dialog));
+		return;
+	}
+	prefs_dialog_open(sensors_applet);
 }
 
 static void about_cb(BonoboUIComponent *uic,
@@ -55,10 +61,24 @@ static void about_cb(BonoboUIComponent *uic,
 		gtk_window_present(GTK_WINDOW(sensors_applet->about_dialog));
 		return;
 	}
-#endif // HAVE_GTK_26
+#endif /* HAVE_GTK_26 */
 	about_dialog_open(sensors_applet);
 }
 
+static void help_cb (BonoboUIComponent *uic, 
+		     SensorsApplet *sensors_applet,
+		     const gchar *verbname) {
+	   GError *error = NULL;
+	   
+	   gnome_help_display(PACKAGE, NULL,
+			      &error);
+
+	   if (error) {
+			 g_printf("Could not open help document: %s\n ",error->message);
+			 g_error_free (error);
+	   }
+}
+
 static void destroy_cb(GtkWidget *widget, SensorsApplet *sensors_applet) {
 	/* destory sensors tree and dialogs and finally the applet */
 	if (sensors_applet->timeout_id) {
@@ -70,7 +90,7 @@ static void destroy_cb(GtkWidget *widget, SensorsApplet *sensors_applet) {
 	}
 
 	if (sensors_applet->prefs_dialog != NULL) {
-		gtk_widget_destroy(GTK_WIDGET(sensors_applet->prefs_dialog->dialog));
+		gtk_widget_destroy(GTK_WIDGET(sensors_applet->prefs_dialog));
 	}
 	gtk_widget_destroy(GTK_WIDGET(sensors_applet->applet));
 
@@ -83,12 +103,11 @@ static void change_background_cb(PanelApplet *applet,
 				 GdkColor *color, 
 				 GdkPixmap *pixmap, 
 				 SensorsApplet *sensors_applet) {
-
-	g_assert(applet == sensors_applet->applet);
-
 	GtkRcStyle *rc_style;
 	GtkStyle *style;
 
+	g_assert(applet == sensors_applet->applet);
+
 	/* reset style */
 	gtk_widget_set_style(GTK_WIDGET(applet), NULL);
 	rc_style = gtk_rc_style_new();
@@ -120,6 +139,7 @@ static void change_background_cb(PanelApplet *applet,
 
 static const BonoboUIVerb sensors_applet_menu_verbs[] = {
 	BONOBO_UI_UNSAFE_VERB("Preferences", prefs_cb),
+	BONOBO_UI_UNSAFE_VERB ("Help", help_cb),
 	BONOBO_UI_UNSAFE_VERB("About", about_cb),
 	BONOBO_UI_VERB_END
 };
@@ -133,29 +153,43 @@ void sensors_applet_register_sensors_interface(SensorsApplet *sensors_applet,
 }
 
 
-/* path should be the full path to a file representing the sensor (eg
- * /dev/hda or /sys/devices/platform/i2c-0/0-0290/temp1_input) */
+   
+  
+
 gboolean sensors_applet_add_sensor_full_details(SensorsApplet *sensors_applet,
-					    const gchar *path, 
-					    const gchar *id, 
-					    const gchar *label, 
-					    SensorInterface interface, 
-					    SensorType type, 
-					    gboolean enable,
-					    gdouble alarm_value,
-					    AlarmType alarm_type,
-					    gboolean alarm_enable,
-					    const gchar *alarm_command,
-					    gint alarm_timeout,
-					    gdouble multiplier,
-					    gdouble offset) {
+						const gchar *path, 
+						const gchar *id, 
+						const gchar *label, 
+						SensorInterface interface, 
+						SensorType type, 
+						gboolean enable,
+						gdouble alarm_value,
+						AlarmType alarm_type,
+						gboolean alarm_enable,
+						const gchar *alarm_command,
+						gint alarm_timeout,
+						gdouble multiplier,
+						gdouble offset,
+						const gchar *icon_filename) {
 
-	g_assert(sensors_applet != NULL);
 					       
 	GtkTreeIter interfaces_iter, sensors_iter;
+	gboolean not_empty_tree;
+
+	/* assign to UNUSED to ensure we dont accidently match an
+	 * interface that doesnt actually exist */
+	guint node_interface = UNUSED;
+	gboolean not_end_of_interfaces = TRUE, interface_exists = FALSE;
+	gboolean not_end_of_sensors = TRUE;
+	gchar *sensor_id;
+	GdkPixbuf *icon;
+	GError *error = NULL;
+
+	g_assert(sensors_applet != NULL);
 
 	/* assume tree is not empty */
-	gboolean not_empty_tree = TRUE;
+	not_empty_tree = TRUE;
+
 
 	if (NULL == sensors_applet->sensors) {
 
@@ -177,24 +211,24 @@ gboolean sensors_applet_add_sensor_full_details(SensorsApplet *sensors_applet,
 									  * timeout
 									  * index */
 							     G_TYPE_DOUBLE, /* multiplier */
-							     G_TYPE_DOUBLE /* offset */);
+							     G_TYPE_DOUBLE, /* offset */
+							     G_TYPE_STRING, /* icon filename */
+							     GDK_TYPE_PIXBUF); /* icon pixbuf */
+		 
+		
 
 		/* we know tree is actually empty since we just created it */
 		not_empty_tree = FALSE;
 	}
-
-	/* assign to UNUSED to ensure we dont accidently match an
-	 * interface that doesnt actually exist */
-	guint node_interface = UNUSED;
-	gboolean not_end_of_interfaces = TRUE, interface_exists = FALSE;
-	gboolean not_end_of_sensors = TRUE;
-	gchar *sensor_id;
-
+	
+	/* search sensor tree for the parent interface to place this
+	 * sensor under */
 	for (not_empty_tree = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(sensors_applet->sensors), &interfaces_iter); not_empty_tree && not_end_of_interfaces && !interface_exists; not_end_of_interfaces = gtk_tree_model_iter_next(GTK_TREE_MODEL(sensors_applet->sensors), &interfaces_iter)) {
 		gtk_tree_model_get(GTK_TREE_MODEL(sensors_applet->sensors), &interfaces_iter,
 				   INTERFACE_COLUMN, &node_interface,
 				   -1);
 		if (interface == node_interface) {
+			/* found interface in tree */
 			interface_exists = TRUE;
 
 			/* now see if this actual sensor already
@@ -206,6 +240,8 @@ gboolean sensors_applet_add_sensor_full_details(SensorsApplet *sensors_applet,
 						   ID_COLUMN, &sensor_id,
 						   -1);
 				if (g_ascii_strcasecmp(sensor_id, id) == 0) {
+					/* sensor already exists so
+					 * dont add a second time */
 					g_free(sensor_id);
 					return FALSE;
 				}
@@ -218,7 +254,7 @@ gboolean sensors_applet_add_sensor_full_details(SensorsApplet *sensors_applet,
 
 
 	if (!interface_exists) {
-		// wasn't able to find interface root node so create it
+		/* wasn't able to find interface root node so create it */
 		gtk_tree_store_append(sensors_applet->sensors,
 				      &interfaces_iter,
 				      NULL);
@@ -232,6 +268,24 @@ gboolean sensors_applet_add_sensor_full_details(SensorsApplet *sensors_applet,
 		
 	}
 
+	/* try to load the icon */
+	icon = gdk_pixbuf_new_from_file_at_size(icon_filename, DEFAULT_ICON_SIZE, DEFAULT_ICON_SIZE, &error);
+	if (error) {
+		 g_warning ("Could not load icon: %s\nTrying with default icon.\n", error->message);
+		 g_error_free(error);
+		 error = NULL;
+
+		 /* try again with default name */
+		 icon_filename = SENSORS_APPLET_ICON;
+		 icon = gdk_pixbuf_new_from_file_at_size(icon_filename, DEFAULT_ICON_SIZE, DEFAULT_ICON_SIZE, &error);
+		 if (error) {
+			 g_warning ("Could not load icon: %s\n", error->message);
+			 g_error_free(error);
+			 error = NULL;
+			 return;
+		 }
+	}
+
 	/* then add sensor as a child under interface node - ie assume
 	 * we either found it or created it - the inteface node that
 	 * is */
@@ -258,18 +312,30 @@ gboolean sensors_applet_add_sensor_full_details(SensorsApplet *sensors_applet,
 			   ALARM_TIMEOUT_INDEX_COLUMN, -1,
 			   MULTIPLIER_COLUMN, multiplier,
 			   OFFSET_COLUMN, offset,
+			   ICON_FILENAME_COLUMN, icon_filename,
+			   ICON_PIXBUF_COLUMN, icon,
 			   -1);
 
+	/* remove reference to icon as tree now has ref */
+	g_object_unref(icon);
+
 }	
+
+/* path should be the full path to a file representing the sensor (eg
+ * /dev/hda or /sys/devices/platform/i2c-0/0-0290/temp1_input) */
 gboolean sensors_applet_add_sensor(SensorsApplet *sensors_applet,
-			       const gchar *path, 
-			       const gchar *id, 
-			       const gchar *label, 
-			       SensorInterface interface, 
-			       gboolean enable,
-			       SensorType type) {
+				   const gchar *path, 
+				   const gchar *id, 
+				   const gchar *label, 
+				   SensorInterface interface, 
+				   gboolean enable,
+				   SensorType type,
+				   const gchar *icon_filename) {
 	
 	g_assert(sensors_applet != NULL);
+	if (icon_filename == NULL) {
+		icon_filename = SENSORS_APPLET_ICON;
+	}
 
 	return sensors_applet_add_sensor_full_details(sensors_applet,
 						      path,
@@ -284,86 +350,117 @@ gboolean sensors_applet_add_sensor(SensorsApplet *sensors_applet,
 						      "",
 						      0,
 						      1.0,
-						      0.0);
+						      0.0,
+						      icon_filename);
 } 
 
-
+	
+	
 /* internal helper functions for updating display etc*/
-/* returns a newly allocated string containing the pango fontsize,
-   defaults to medium if error */
-static gchar *sensors_applet_get_font_size(SensorsApplet *sensors_applet) {
+/* returns a constant string containing the pango fontsize, defaults
+   to medium if error - this value is owned and cannot be freed by the
+   caller */
+static const gchar *sensors_applet_get_font_size(SensorsApplet *sensors_applet) {
 	FontSize font_size;
 
 	font_size = panel_applet_gconf_get_int(sensors_applet->applet, FONT_SIZE, NULL);
 	switch (font_size) {
 	case XX_LARGE:
-		return g_strdup(XX_LARGE_TEXT);
+		return XX_LARGE_TEXT;
 	case X_LARGE:
-		return g_strdup(X_LARGE_TEXT);
+		return X_LARGE_TEXT;
 	case LARGE:
-		return g_strdup(LARGE_TEXT);
+		return LARGE_TEXT;
 	case MEDIUM:
-		return g_strdup(MEDIUM_TEXT);
+		return MEDIUM_TEXT;
 	case SMALL:
-		return g_strdup(SMALL_TEXT);
+		return SMALL_TEXT;
 	case X_SMALL:
-		return g_strdup(X_SMALL_TEXT);
+		return X_SMALL_TEXT;
 	case XX_SMALL:
-		return g_strdup(XX_SMALL_TEXT);
+		return XX_SMALL_TEXT;
 	default:
-		return g_strdup(MEDIUM_TEXT);
+		return MEDIUM_TEXT;
 	}
 }
 
-/* HELPER FUNCTIONS for managing labels and values lists */
-/* removes remainder of list links and frees widgets inside list */
-static void sensors_applet_crop_list(GSList **list, GSList **current_element) {
-	while (*current_element != NULL) {
-		GSList *old_element = *current_element;
-		gtk_widget_destroy(GTK_WIDGET(old_element->data));
-		*list = g_slist_remove_link(*list, old_element);
-		*current_element = g_slist_next(*current_element);
-	}
-}
-
-/* take the given text and add if there is already a node at the given
- * element, change its text to this, or create a new node and point
- * current_element at the newly created node */
-static void sensors_applet_add_label_to_list(GSList **list, GSList **current_element, const gchar *text) {
-	
-	GtkLabel *label;
-
-	/* if have reached end of list, need to create a new node etc */
-	if (*current_element == NULL) {
-		/* create a new label and add it to
-		   the list and the display */
-		label = g_object_new(GTK_TYPE_LABEL,
-				     "label", text,
-				     "use-markup", TRUE,
-				     NULL);
-
-		*list = g_slist_append(*list, label);
-
-		/* if we have just created the first
-		   node in the list, we will want to be
-		   pointing at it now */
-		if (g_slist_length(*list) == 1) {
-			*current_element = *list;
-		} else {
-			/* if we just appended a new
-			   node, we will want to be
-			   pointing at this as the
-			   current node */
-			*current_element = g_slist_next(*current_element);
+ /* HELPER FUNCTIONS for managing labels and values lists */ 	 
+ /* removes remainder of list links and frees widgets inside list */ 	 
+ static void sensors_applet_crop_list(GSList **list, GSList **current_element) { 	 
+         while (*current_element != NULL) { 	 
+                 GSList *old_element = *current_element; 	 
+                 gtk_widget_destroy(GTK_WIDGET(old_element->data)); 	 
+                 *list = g_slist_remove_link(*list, old_element); 	 
+                 *current_element = g_slist_next(*current_element); 	 
+         } 	 
+ } 	 
+  	 
+ /* take the given text and add if there is already a node at the
+  * given element, change its text to this, or create a new node and
+  * point current_element at the newly created node */ 	 
+static void sensors_applet_add_label_to_list(GSList **list, GSList **current_element, const gchar *text) { 	 
+	/* if have reached end of list, need to create a new node etc */ 	 
+	if (*current_element == NULL) { 	 
+		/* create a new label and add it to 	 
+		   the list and the display */ 	 
+		*list = g_slist_append(*list, g_object_new(GTK_TYPE_LABEL,
+							   "use-markup", TRUE,
+							   "label", text,
+							   NULL)); 	 
+  	 
+		/* if we have just created the first 	 
+		   node in the list, we will want to be 	 
+		   pointing at it now */ 	 
+		if (g_slist_length(*list) == 1) { 	 
+			*current_element = *list; 	 
+		} else { 	 
+			/* if we just appended a new 	 
+			   node, we will want to be 	 
+			   pointing at this as the 	 
+			   current node */ 	 
+			*current_element = g_slist_next(*current_element); 	 
 			
-		}
+		} 	 
 	} else {
-		/* assign text to current element (a GtkLabel) */
-		gtk_label_set_markup(GTK_LABEL((*current_element)->data), text);
-	}
+		  gtk_label_set_label(GTK_LABEL((*current_element)->data), text);
+	} 	 
 }
 
-gboolean sensors_applet_execute_alarm(const gchar *alarm_command) {
+static void sensors_applet_add_image_to_list(GSList **list, GSList **current_element, GdkPixbuf *pixbuf) { 	 
+
+	GtkWidget *image;
+         /* if have reached end of list, need to create a new node etc */ 	 
+	if (*current_element == NULL) {
+		/* create a new image and add it to 	 
+		   the list and the display */ 	 
+		 image = gtk_image_new_from_pixbuf(pixbuf);
+                 *list = g_slist_append(*list, image); 	 
+  	 
+                 /* if we have just created the first 	 
+                    node in the list, we will want to be 	 
+                    pointing at it now */ 	 
+                 if (g_slist_length(*list) == 1) { 	 
+                         *current_element = *list; 	 
+                 } else { 	 
+                         /* if we just appended a new 	 
+                            node, we will want to be 	 
+                            pointing at this as the 	 
+                            current node */ 	 
+                         *current_element = g_slist_next(*current_element); 	 
+  	 
+                 } 	 
+         } else { 	 
+		 /* and set to new one if needed */
+		 if (gtk_image_get_pixbuf(GTK_IMAGE((*current_element)->data)) != pixbuf) {
+			 gtk_image_set_from_pixbuf(GTK_IMAGE((*current_element)->data),
+						   pixbuf);
+		 }
+		 
+		 
+         } 	 
+}
+
+static gboolean sensors_applet_execute_alarm(const gchar *alarm_command) {
 	gnome_execute_shell(NULL, alarm_command);
 	return TRUE;
 }
@@ -377,7 +474,7 @@ static guint sensors_applet_add_alarm(SensorsApplet *sensors_applet, const gchar
 }
 
 static void sensors_applet_alarm_on(SensorsApplet *sensors_applet, GtkTreeIter *iter) {
-	static gchar *alarm_command;
+	gchar *alarm_command;
 	gint alarm_timeout_index;
 	guint alarm_timeout;
 	gtk_tree_model_get(GTK_TREE_MODEL(sensors_applet->sensors),
@@ -393,8 +490,8 @@ static void sensors_applet_alarm_on(SensorsApplet *sensors_applet, GtkTreeIter *
 				   ALARM_COMMAND_COLUMN, &alarm_command,
 				   ALARM_TIMEOUT_COLUMN, &alarm_timeout,
 				   -1);
-		// execute alarm once, then add to time to
-		// keep repeating it
+		/* execute alarm once, then add to time to
+		   keep repeating it */
 		gnome_execute_shell(NULL, alarm_command);
 		if (alarm_timeout <= 0) {
 			alarm_timeout_index = sensors_applet_add_alarm(sensors_applet, alarm_command, G_MAXINT);
@@ -414,7 +511,6 @@ static void sensors_applet_alarm_on(SensorsApplet *sensors_applet, GtkTreeIter *
 void sensors_applet_alarm_off(SensorsApplet *sensors_applet, GtkTreeIter *iter) {
 	gint alarm_timeout_index;
 	gchar *alarm_command;
-	gchar *temp;
 	GSList *current_alarm_command;
 
 	gtk_tree_model_get(GTK_TREE_MODEL(sensors_applet->sensors),
@@ -448,15 +544,44 @@ void sensors_applet_alarm_off(SensorsApplet *sensors_applet, GtkTreeIter *iter)
 	}
 }
 
-void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
+/* should be called as a g_container_foreach at the start of
+ * pack_display if ythe table already exists to remove but keep alive
+ * all children of the table before repacking it */
+static void sensors_applet_pack_display_empty_table_cb(GtkWidget *widget,
+						   gpointer data) {
+	GtkContainer *container = GTK_CONTAINER(data);
+
+	/* ref then remove widget */
+	g_object_ref(widget);
+	gtk_container_remove(container, widget);
+}
+
+/* should be called as a g_container_foreach at the end of
+ * pack_display to unref any of the old children that we have readdded
+ * to the table to stop reference creep from the g_object_ref called
+ * on each child at the start of pack labels */
+static void sensors_applet_pack_display_cleanup_refs_cb(GtkWidget *widget,
+							gpointer data) {
+						  
+	GList *old_children = (GList *)data;
+	if (g_list_find(old_children, widget)) {
+		g_object_unref(widget);
+	}
+}
+						 
+static void sensors_applet_pack_display(SensorsApplet *sensors_applet) {
 	static GtkLabel *no_sensors_enabled_label = NULL;
 	int num_active_sensors = 0, num_sensors_per_group, cols, rows, i, j;
+	GList *old_table_children = NULL;
+	GList *current_child;
+
 	GSList *current_value;
 	GSList *current_label;
+	GSList *current_icon;
+	gint display_mode;
 
-	/* get total number of active sensors - this is the number of
-	   values labels that have been created */
 	num_active_sensors = g_slist_length(sensors_applet->values);
+	display_mode = panel_applet_gconf_get_int(sensors_applet->applet, DISPLAY_MODE, NULL);
 
 	num_sensors_per_group = panel_applet_gconf_get_int(sensors_applet->applet, NUM_SENSORS, NULL);
 
@@ -484,17 +609,17 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 
 	/* if displaying labels need to modify number of rows / colums
 	   to accomodate this */
-	if (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_LABELS, NULL)) {
-		if (panel_applet_gconf_get_bool(sensors_applet->applet, LABELS_INLINE, NULL)) {
-			/* to display labels next to values need twice
-			   as many columns */
-			cols *= 2;
-		} else {
-			/* to display labels above values, we need
-			 * twice as many rows as without */
-			rows *= 2;
-		}
-	}
+	 if (display_mode != DISPLAY_NONE) {
+		 if (panel_applet_gconf_get_bool(sensors_applet->applet, LABELS_INLINE, NULL)) {
+			 /* to display labels next to values need twice
+			    as many columns */
+			 cols *= 2;
+		 } else {
+			 /* to display labels above values, we need
+			  * twice as many rows as without */
+			 rows *= 2;
+		 }
+	 }	 
 
 	if (sensors_applet->table == NULL) {
 		/* create table and add to applet */
@@ -505,17 +630,18 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 	} else {
 		/* remove all children if table already exists so we can start
 		 * again */
-		GList *children, *current_child;
-		children = gtk_container_get_children(GTK_CONTAINER(sensors_applet->table));
-		for (current_child = g_list_first(children); current_child != NULL; current_child = g_list_next(current_child)) {
-			/* ensure widget does not get destroyed */
-			g_object_ref(current_child->data);
-			gtk_container_remove(GTK_CONTAINER(sensors_applet->table), GTK_WIDGET(current_child->data));
-		}
+		/* save a list of the old children for later */
+		old_table_children = gtk_container_get_children(GTK_CONTAINER(sensors_applet->table));
+
+		gtk_container_foreach(GTK_CONTAINER(sensors_applet->table),
+				      sensors_applet_pack_display_empty_table_cb,
+				      sensors_applet->table);
+
 		/* then resize table */
 		gtk_table_resize(GTK_TABLE(sensors_applet->table), rows, cols);
 	}
 	
+	/* if no active sensors, display a label to this effect and return */
 	if (num_active_sensors == 0) {
 		if (no_sensors_enabled_label == NULL) {
 			no_sensors_enabled_label = g_object_new(GTK_TYPE_LABEL,
@@ -526,37 +652,45 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 					  GTK_WIDGET(no_sensors_enabled_label),
 					  0, 1,
 					  0, 1);
-		return;
-		
+
 	}
 
 	current_value = sensors_applet->values;
 	current_label = sensors_applet->labels;
+	current_icon = sensors_applet->icons;
 
-	/* pack labels and values into table */
+	/* pack icons / labels and values into table */
 	
 	/* if showing labels need to pack these first */
-	if (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_LABELS, NULL)) {
-		/* need to first pack labels, then the values either
-		   go under the labels or next to them */
-		
+	if (display_mode != DISPLAY_NONE) {
 		/* loop through columns */
-		for (i = 0; current_label != NULL && current_value != NULL && i < cols; /* increments depends on how we lay them out - see below */) {
+		for (i = 0; current_label != NULL && current_value != NULL && current_icon != NULL && i < cols; /* increments depends on how we lay them out - see below */) {
 			
 			/* loop through rows in a column */
-			for (j = 0; current_value != NULL && current_label != NULL && j < rows; /* see bottom of for loop*/) {
-				/* attach label at this point */
-				gtk_table_attach_defaults(GTK_TABLE(sensors_applet->table),
-							  GTK_WIDGET(current_label->data),
+			for (j = 0; current_label != NULL && current_value != NULL && current_icon != NULL && j < rows; /* see bottom of for loop*/) {
+				/* attach label / icon at this point */
+				if (display_mode == DISPLAY_ICON) {
+					gtk_table_attach_defaults(GTK_TABLE(sensors_applet->table),
+								  GTK_WIDGET(current_icon->data),
 							  i, i + 1,
 							  j, j + 1);
-				
+				} else {
+					gtk_table_attach_defaults(GTK_TABLE(sensors_applet->table),
+								  GTK_WIDGET(current_label->data),
+							  i, i + 1,
+							  j, j + 1);
+				}				
 				/* now attach sensor value to either
 				   row below or column next to */
-				if (panel_applet_gconf_get_bool(sensors_applet->applet, LABELS_INLINE, NULL)) { /* place value next to label */
+				if (panel_applet_gconf_get_bool(sensors_applet->applet, LABELS_INLINE, NULL)) { 
 					/* left align labels */
-					gtk_misc_set_alignment(GTK_MISC(current_label->data), 0.0, 0.5);
+                                        gtk_misc_set_alignment(GTK_MISC(current_icon->data), 0.0, 0.5);
+					gtk_misc_set_alignment(GTK_MISC(current_label->data), 0.0, 0.5); 	 
 					gtk_misc_set_alignment(GTK_MISC(current_value->data), 0.0, 0.5);
+ 
+
+
+					 /* place value next to label */
 					gtk_table_attach_defaults(GTK_TABLE(sensors_applet->table),
 								  GTK_WIDGET(current_value->data),
 								  i + 1, i + 2,
@@ -564,10 +698,11 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 					
 					j++;
 				} else { /* place value below label */
-					/* center align labels */
-					gtk_misc_set_alignment(GTK_MISC(current_label->data), 0.5, 0.5);
-					gtk_misc_set_alignment(GTK_MISC(current_value->data), 0.5, 0.5);
-
+					/* center align labels */ 	 
+					gtk_misc_set_alignment(GTK_MISC(current_icon->data), 0.5, 0.5); 	 
+					gtk_misc_set_alignment(GTK_MISC(current_label->data), 0.5, 0.5); 	 
+					gtk_misc_set_alignment(GTK_MISC(current_value->data), 0.5, 0.5); 	 
+ 
 					gtk_table_attach_defaults(GTK_TABLE(sensors_applet->table),
 								  GTK_WIDGET(current_value->data),
 								  i, i + 1,
@@ -575,9 +710,10 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 					j += 2;
 				}
 				current_label = g_slist_next(current_label);
+				current_icon = g_slist_next(current_icon);
 				current_value = g_slist_next(current_value);
 
-			} // end row loop
+			} /* end row loop */
 			/* now increment column index as needed */
 			if (panel_applet_gconf_get_bool(sensors_applet->applet, LABELS_INLINE, NULL)) { /* place value next to label */
 				i += 2;
@@ -586,10 +722,12 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 			}
 			
 			
-		} // end column loop	
-	} else { /* not showing labels so just pack values */
+		} /* end column loop	*/
+
+		
+	} else { /* not showing labels or icons so just pack values */
 		for (i = 0; current_value != NULL && i < cols; ++i) {
-			for (j = 0; current_value != NULL && j < rows; ++j) {
+			for (j = 0; current_value!= NULL && j < rows; ++j) {
 				gtk_table_attach_defaults(GTK_TABLE(sensors_applet->table),
 							  GTK_WIDGET(current_value->data),
 							  i, i + 1,
@@ -600,19 +738,29 @@ void sensors_applet_pack_labels(SensorsApplet *sensors_applet) {
 		}
 		
 	}
-	
+	if (old_table_children != NULL) {
+		gtk_container_foreach(GTK_CONTAINER(sensors_applet->table),
+				      sensors_applet_pack_display_cleanup_refs_cb,
+				      old_table_children);
+		g_list_free(old_table_children);
+	}
+
 } 	    
 
-void sensors_applet_update_display(SensorsApplet *sensors_applet) {
+static void sensors_applet_update_display_only(SensorsApplet *sensors_applet) {
+	sensors_applet_update_display(sensors_applet,
+				      FALSE);
+}
+
+void sensors_applet_update_display(SensorsApplet *sensors_applet,
+				   gboolean repack_display) {
 	/* walk through tree structure and for each enabled sensor
 	 * call it's get_sensor_value function and then update labels
 	 * etc */
 
-	g_assert(sensors_applet->sensors != NULL);
-
-	static gint last_number_of_active_sensors = 0;
 	gchar *tooltip;
 	gint number_of_active_sensors = 0;
+	static gint last_number_of_active_sensors = 0;
 
 	/* to iterate through the sensors data structure */
 	GtkTreeIter interfaces_iter, sensors_iter;
@@ -629,19 +777,27 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 	AlarmType sensor_alarm_type;
 	gdouble sensor_multiplier;
 	gdouble sensor_offset;
-	gfloat sensor_value;
+	gdouble sensor_value;
 
 	/* to build the list of labels as we go */
 	gchar *label_text = NULL;
 	gchar *value_text = NULL;
-	gchar *font_size = NULL;
+	const gchar *font_size;
+	gchar *old_value_text;
+	GdkPixbuf *icon;
 
 	GSList *current_label = NULL;
 	GSList *current_value = NULL;
+	GSList *current_icon = NULL;
+	
+	GError *error = NULL;
+
+	g_assert(sensors_applet->sensors != NULL);
 
 	/* get the first label in the list */
 	current_label = sensors_applet->labels;
 	current_value = sensors_applet->values;
+	current_icon = sensors_applet->icons;
 	
 	/* now step through the GtkTreeStore sensors to
 	   find which sensors are enabled */
@@ -662,32 +818,33 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 					   ALARM_ENABLE_COLUMN, &sensor_alarm_enabled,
 					   MULTIPLIER_COLUMN, &sensor_multiplier,
 					   OFFSET_COLUMN, &sensor_offset,
+					   ICON_PIXBUF_COLUMN, &icon,
 					   -1);
 			
 			if (sensor_enabled) {
 				/* get font size */
 				font_size = sensors_applet_get_font_size(sensors_applet);
 				
-				GError *error = NULL;
 				sensor_value = sensors_applet->get_sensor_value[sensor_interface](sensor_path,
 												  sensor_id,
 												  sensor_type,
 												  &error);
 				
 				
-				if (error != NULL) {
+				if (error) {
 					value_text = g_strdup(error->message);
 					g_error_free(error);
+					error = NULL;
 				} else { 
 					switch (sensor_type) {
 					case TEMP_SENSOR: 
-						if (panel_applet_gconf_get_bool(sensors_applet->applet, FAHRENHEIT, NULL)) {
+						if (panel_applet_gconf_get_bool(sensors_applet->applet, FARENHEIT, NULL)) {
 							sensor_value = (9.0 * sensor_value / 5.0) + 32.0;
 							sensor_value = (sensor_value * sensor_multiplier) + sensor_offset;
-							value_text = g_strdup_printf("%3.1f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_UNITS, NULL) ? "\302\260F" : ""));
+							value_text = g_strdup_printf("%2.0f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_UNITS, NULL) ? "\302\260F" : ""));
 						} else {
 							sensor_value = (sensor_value * sensor_multiplier) + sensor_offset;
-							value_text = g_strdup_printf("%3.1f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_UNITS, NULL) ? "\302\260C" : ""));
+							value_text = g_strdup_printf("%2.0f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_UNITS, NULL) ? "\302\260C" : ""));
 						}
 						break;
 						
@@ -696,7 +853,6 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 						value_text = g_strdup_printf("%4.0f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_UNITS, NULL) ? _("RPM") : ""));
 						break;
 						
-					case CPU_VID_SENSOR: /* fall through to voltage sensor */
 					case VOLTAGE_SENSOR:
 						sensor_value = (sensor_value * sensor_multiplier) + sensor_offset;
 						value_text = g_strdup_printf("%4.2f%s", sensor_value, (panel_applet_gconf_get_bool(sensors_applet->applet, SHOW_UNITS, NULL) ? _("V") : ""));
@@ -710,8 +866,8 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 					default:
 						g_assert_not_reached();	
 						
-					} // end switch(sensor_type)
-					gchar *old_value_text = value_text;
+					} /* end switch(sensor_type) */
+				        old_value_text = value_text;
 					
 					if (sensor_alarm_enabled) {
 						switch (sensor_alarm_type) {
@@ -747,7 +903,9 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 							break;
 						default: 
 							g_assert_not_reached();
-						} // end switch(sensor_alarm_type)
+						} /* end
+						   * switch(sensor_alarm_type) */
+
 					} else {
 						/* just do font size */
 						value_text = g_markup_printf_escaped("<span size=\"%s\">%s</span>", font_size, old_value_text);
@@ -768,28 +926,36 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 				sensors_applet_add_label_to_list(&(sensors_applet->labels), &current_label, label_text);
 				g_free(label_text);
 				current_label = g_slist_next(current_label);
+
+				/* add icon to icons list */
+				sensors_applet_add_image_to_list(&(sensors_applet->icons), &current_icon, icon);
+				/* g_object_unref(icon) */
+				current_icon = g_slist_next(current_icon);
 				
-				/* free allocated string from font size */
-				g_free(font_size);
-			} // end if (sensor_enabled)
+			} /* end if (sensor_enabled) */
 			g_free(sensor_path);
+			g_free(sensor_id);
 			g_free(sensor_label);
-		} // end for loop of sensors
-	} // end for loop of interfaces
+			g_object_unref(icon);
+
+		} /* end for loop of sensors */
+	} /* end for loop of interfaces */
 	
 	/* now remove any remaining elements in either list to free up
 	   memory */
 	sensors_applet_crop_list(&(sensors_applet->values), &current_value);
 	sensors_applet_crop_list(&(sensors_applet->labels), &current_label);
+	sensors_applet_crop_list(&(sensors_applet->icons), &current_icon);
 	
 	number_of_active_sensors = g_slist_length(sensors_applet->values);
 
-	if (number_of_active_sensors != last_number_of_active_sensors) {
-		/* now repack labels into table update display of applets */
-		sensors_applet_pack_labels(sensors_applet);
+	/* now repack labels into table update display of applets */
+	if (repack_display || last_number_of_active_sensors != number_of_active_sensors) {
+		sensors_applet_pack_display(sensors_applet);
 	}
 	last_number_of_active_sensors = number_of_active_sensors;
 
+
 	tooltip = g_strdup_printf("%s\n%d %s",
 				  _("Sensors Applet"),
 				  number_of_active_sensors,
@@ -806,9 +972,9 @@ void sensors_applet_update_display(SensorsApplet *sensors_applet) {
 }
 
 static void sensors_applet_write_defaults_to_gconf(SensorsApplet *sensors_applet) {
-	panel_applet_gconf_set_bool(sensors_applet->applet, FAHRENHEIT, FALSE, NULL);
+	panel_applet_gconf_set_bool(sensors_applet->applet, FARENHEIT, FALSE, NULL);
 	panel_applet_gconf_set_int(sensors_applet->applet, NUM_SENSORS, 1, NULL);
-	panel_applet_gconf_set_bool(sensors_applet->applet, SHOW_LABELS, TRUE, NULL);
+	panel_applet_gconf_set_int(sensors_applet->applet, DISPLAY_MODE, DISPLAY_LABEL, NULL);
 	panel_applet_gconf_set_bool(sensors_applet->applet, LABELS_INLINE, TRUE, NULL);
 	panel_applet_gconf_set_bool(sensors_applet->applet, SHOW_UNITS, TRUE, NULL);
 	panel_applet_gconf_set_int(sensors_applet->applet, TIMEOUT, DEFAULT_TIMEOUT, NULL);
@@ -823,6 +989,7 @@ static void sensors_applet_setup_sensors_interfaces(SensorsApplet *sensors_apple
 	i2c_proc_sensors_interface_init(sensors_applet);
 	i2c_sys_sensors_interface_init(sensors_applet);
 	i8k_sensors_interface_init(sensors_applet);
+	ibm_acpi_sensors_interface_init(sensors_applet);
 	omnibook_sensors_interface_init(sensors_applet);
 	pmu_sys_sensors_interface_init(sensors_applet);
 }
@@ -869,22 +1036,15 @@ void sensors_applet_init(SensorsApplet *sensors_applet) {
 		label = gtk_label_new(_("No sensors found!"));
 		gtk_container_add(GTK_CONTAINER(sensors_applet->applet), label);
 		gtk_widget_show_all(GTK_WIDGET(sensors_applet->applet));
-		sensors_applet->prefs_dialog = g_new0(PrefsDialog, 1);
-		sensors_applet->prefs_dialog->sensors_applet = sensors_applet;
-		prefs_dialog_init(sensors_applet->prefs_dialog);
 		return;
 	}
 	
-	sensors_applet->prefs_dialog = g_new0(PrefsDialog, 1);
-	sensors_applet->prefs_dialog->sensors_applet = sensors_applet;
-	prefs_dialog_init(sensors_applet->prefs_dialog);
 
 	sensors_applet->tooltips = gtk_tooltips_new();
 	gtk_tooltips_set_tip(sensors_applet->tooltips, GTK_WIDGET(sensors_applet->applet), _("Sensors Applet"), "");
-	sensors_applet_update_display(sensors_applet);
+	sensors_applet_update_display(sensors_applet, TRUE);
 
-	sensors_applet_pack_labels(sensors_applet);
-	sensors_applet->timeout_id = g_timeout_add(panel_applet_gconf_get_int(sensors_applet->applet, TIMEOUT, NULL), (GSourceFunc)sensors_applet_update_display, sensors_applet);
+	sensors_applet->timeout_id = g_timeout_add(panel_applet_gconf_get_int(sensors_applet->applet, TIMEOUT, NULL), (GSourceFunc)sensors_applet_update_display_only, sensors_applet);
 
 	gtk_widget_show_all(GTK_WIDGET(sensors_applet->applet));
 }
